/* Tree-dumping functionality for intermediate representation.
   Copyright (C) 1999, 2000 Free Software Foundation, Inc.
   Written by Mark Mitchell <mark@codesourcery.com>

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include "config.h"
#include "system.h"
#include "tree.h"
#include "cp-tree.h"
#include "c-dump.h"

#include "gd_graph.h"

extern gd_graph the_graph;

int
cp_dump_tree (di, t)
     dump_info_p di;
     tree t;
{
  enum tree_code code;

  if (DECL_P (t)) {
    /* note language of declaration */
    if (DECL_LANG_SPECIFIC(t)){
      switch (DECL_LANGUAGE(t)) {
	case lang_c:
	  the_graph.nodeBuffer[the_graph.curNodePos].decl.lang =
	    src_lang_c;
	  break;
	case lang_cplusplus:
	  the_graph.nodeBuffer[the_graph.curNodePos].decl.lang =
	    src_lang_cp;
	  break;
	case lang_java:
	  the_graph.nodeBuffer[the_graph.curNodePos].decl.lang =
	    src_lang_java;
	  break;
	default:
	  fprintf(stderr,"unknown language %d\n", DECL_LANGUAGE(t));
	  exit(1);
      }
    } else {
	the_graph.nodeBuffer[the_graph.curNodePos].decl.lang =
	    src_lang_none;
    }
  }

  /* Figure out what kind of node this is.  */
  code = TREE_CODE (t);
  switch (code) {
    case NAMESPACE_DECL:
      /* The fake `::std' namespace does not have DECL_LANG_SPECIFIC,
	 and therefore many other macros do not work on it.  */
      if (t == fake_std_node) {
	fprintf(stderr,"fake std node??\n");
	break;
      }
      if (DECL_NAMESPACE_ALIAS (t)) {
	the_graph.nodeBuffer[the_graph.curNodePos].namespace_decl.dcls = -1;
	the_graph.nodeBuffer[the_graph.curNodePos].namespace_decl.alis =
	  queue_and_return_index (di,DECL_NAMESPACE_ALIAS (t),DUMP_NONE)-1;
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].namespace_decl.alis = -1;
	the_graph.nodeBuffer[the_graph.curNodePos].namespace_decl.dcls =
	  queue_and_return_index (di,cp_namespace_decls (t),DUMP_NONE)-1;
      }
      break;

    case FUNCTION_DECL:
      if (!DECL_THUNK_P (t)) {
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.member =
	  (DECL_FUNCTION_MEMBER_P (t))?1:0;
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.constr =
	  (DECL_CONSTRUCTOR_P (t))?1:0;
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.decon =
	  (DECL_DESTRUCTOR_P (t))?1:0;
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.oper =
	  (DECL_OVERLOADED_OPERATOR_P (t))?1:0;
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.conver =
	  (DECL_CONV_FN_P (t))?1:0;
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.ginit =
	  (DECL_GLOBAL_CTOR_P (t))?1:0;
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.gfini =
	  (DECL_GLOBAL_DTOR_P (t))?1:0;
      }
      break;
    case VAR_DECL:
    case PARM_DECL:
    case TYPE_DECL:
    case LABEL_DECL:
    case CONST_DECL:
    case RESULT_DECL:
    case FIELD_DECL:
      break;
    case TEMPLATE_DECL:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTEMPLATE_DECL;
      the_graph.nodeBuffer[the_graph.curNodePos].template_decl.rslt =
	queue_and_return_index (di,DECL_TEMPLATE_RESULT(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].template_decl.inst =
	queue_and_return_index (di,DECL_TEMPLATE_INSTANTIATIONS(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].template_decl.spcs =
	queue_and_return_index (di,DECL_TEMPLATE_SPECIALIZATIONS(t),DUMP_NONE)-1;
      /* template parameters */
      /* because this uses a tree_vec node, which is variable length
       * and gd does not support variable length nodes, we translate
       * to a list (just like in the inheritance in record and union types).
       * It is a bit easier here since each element of the vector points
       * to elements of a linked list. So we just point at the beginning
       * of the linked list. This area is not well documented on gcc.gnu.org
       * so we include much verification code so that if we run into
       * graphs that don't match our expectation, we can notify the user */
      /* we are expecting
       *   -parms-> tree_list
       *             -purpose-> int_cst(1)
       *             -value-> tree_vec
       *                        -0-> tree_list ->-chan->-|
       *                                  V-------<----<-|
       *                                  V
       *                        -1-> tree_list ->-chan->-|
       *                                  V-------<----<-|
       *                                  V
       *                        -2-> tree_list
       * where the value ptr for each of the tree_list nodes
       * point to the declarations of the template paramters.
       * So we will point the parms field directly at tree_vec(0)
       */
      {
	tree parms,cst1,vec;
	int i;

	parms = DECL_TEMPLATE_PARMS(t);
	/* verify that this is a tree_list */
	if (TREE_CODE(parms) != TREE_LIST) {
	  fprintf(stderr,"\nparms of %ld(template_decl) is not a tree_list\n",
	      the_graph.curNodePos);
	  fprintf(stderr," Its code: %d(%s)",
	      TREE_CODE(parms),
	      tree_code_name[(int)TREE_CODE(parms)]);
	  exit(1);
	}
	/* verify nothing else in list */
	if (TREE_CHAIN(parms) != NULL) {
	  fprintf(stderr,"\nparms of %ld(template_decl) is not a single tree_list\n",
	      the_graph.curNodePos);
	  fprintf(stderr,"it has more than one node!!\n");
	  exit(1);
	}
	cst1 = TREE_PURPOSE(parms);
	/* verify that cst1 is an integercst with value 1 */
	if (TREE_CODE(cst1) != INTEGER_CST){
	  fprintf(stderr,"\npurpose of %ld(template_decl) is not integer_cst\n",
	      the_graph.curNodePos);
	  fprintf(stderr," Its code: %d(%s)",
	      TREE_CODE(cst1),
	      tree_code_name[(int)TREE_CODE(cst1)]);
	  exit(1);
	}
	if (TREE_INT_CST_HIGH(cst1) != 0 &&
	    TREE_INT_CST_LOW(cst1) != 1){
	  fprintf(stderr,"\npurpose %ld(template_decl) is not value 1\n",

	      the_graph.curNodePos);
	  fprintf(stderr,"value = %d,%d\n",TREE_INT_CST_HIGH(cst1),
	      TREE_INT_CST_LOW(cst1));
	  exit(1);
	}
	vec = TREE_VALUE(parms);
	/* verify that this is a tree_vec */
	if (TREE_CODE(vec) != TREE_VEC){
	  fprintf(stderr,"\nvalue of %ld(template_decl) is not tree_vec\n",
	      the_graph.curNodePos);
	  fprintf(stderr," Its code: %d(%s)",
	      TREE_CODE(vec),
	      tree_code_name[(int)TREE_CODE(vec)]);
	  exit(1);
	}
	/* verify that each element of the list points to a tree_list */
	for(i = 0; i < TREE_VEC_LENGTH(vec); i++){
	  tree element;
	  element = TREE_VEC_ELT(vec,i);
	  if (TREE_CODE(element) != TREE_LIST){
	    fprintf(stderr,"tree vec of template_decl(%ld) is not tree list\n",
		the_graph.curNodePos);
	  fprintf(stderr," Its code: %d(%s)",
	      TREE_CODE(element),
	      tree_code_name[(int)TREE_CODE(element)]);
	    exit(1);
	  }
	}
	/* verify tree list linked together */
	for(i = 0; i < TREE_VEC_LENGTH(vec)-1; i++){
	  if(TREE_CHAIN(TREE_VEC_ELT(vec,i)) != TREE_VEC_ELT(vec,i+1)){
	    fprintf(stderr,"tree_list of tree vec of template %ld is not linked(%d)\n",
		the_graph.curNodePos,i);
	    exit(1);
	  }
	}
	the_graph.nodeBuffer[the_graph.curNodePos].template_decl.parms =
	  queue_and_return_index(di,TREE_VEC_ELT(vec,0),DUMP_NONE)-1;
      } 
      break;
    case USING_DECL:
      break;
    /* types */
    case FUNCTION_TYPE:
    case INTEGER_TYPE:
    case ARRAY_TYPE:
    case VOID_TYPE:
    case REAL_TYPE:
    case COMPLEX_TYPE:
    case VECTOR_TYPE:
    case ENUMERAL_TYPE:
    case BOOLEAN_TYPE:
    case CHAR_TYPE:
      break;
    case POINTER_TYPE:
      if (TYPE_PTRMEM_P (t)) {
	the_graph.nodeBuffer[the_graph.curNodePos].pointer_type.ptrmem = 1;
	the_graph.nodeBuffer[the_graph.curNodePos].pointer_type.ptd =
	  queue_and_return_index(di,TYPE_PTRMEM_POINTED_TO_TYPE(t),DUMP_NONE)-1;
	the_graph.nodeBuffer[the_graph.curNodePos].pointer_type.cls =
	  queue_and_return_index(di,TYPE_PTRMEM_CLASS_TYPE(t),DUMP_NONE)-1;
	return 1;
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].pointer_type.ptrmem = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].pointer_type.ptd = -1;
	the_graph.nodeBuffer[the_graph.curNodePos].pointer_type.cls = -1;
      }
      break;
    case OFFSET_TYPE:
    case REFERENCE_TYPE:
    case METHOD_TYPE:
    case FILE_TYPE:
    case SET_TYPE:
      break;
    case RECORD_TYPE:
      if (TYPE_PTRMEMFUNC_P(t)){
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.ptrmem = 1;
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.ptd =
	  queue_and_return_index (di,TYPE_PTRMEM_POINTED_TO_TYPE(t),DUMP_NONE)-1;
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.cls =
	  queue_and_return_index (di,TYPE_PTRMEM_CLASS_TYPE(t),DUMP_NONE)-1;
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.vfld = -1;
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.ptrmem = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.ptd = -1;
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.cls = -1;
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.vfld = 
	  queue_and_return_index (di,TYPE_VFIELD(t),DUMP_NONE)-1;
      }
      the_graph.nodeBuffer[the_graph.curNodePos].record_type.isClass =
	CLASSTYPE_DECLARED_CLASS(t);
      /* gcc uses variable size nodes here. GD does not have variable
       * size nodes, so we transform the vector into a list */
      {
	/* number of base classes */
	int n_baseclasses = CLASSTYPE_N_BASECLASSES(t);
	if (n_baseclasses != 0) {
	  tree binfos = TYPE_BINFO_BASETYPES(t);
	  tree tmpnodes;
	  int i;
	  tmpnodes = malloc((unsigned)(sizeof(union tree_node) * n_baseclasses));
	  if (tmpnodes == NULL) {
	    fprintf(stderr,"could not allocate extra nodes for inheritance\n");
	    exit(1);
	  }
	  /* fill array */
	  for(i = 0; i < n_baseclasses; i++) {
	    tree base_binfo = TREE_VEC_ELT(binfos,i);
	    tmpnodes[i].common.code=TREE_LIST;
	    tmpnodes[i].common.type=NULL;
	    if (i != n_baseclasses -1){
	      tmpnodes[i].common.chain=&(tmpnodes[i+1]);
	    } else{
	      tmpnodes[i].common.chain=NULL;
	    }
	    tmpnodes[i].list.purpose=NULL;
	    tmpnodes[i].list.value= BINFO_TYPE(base_binfo);
	    tmpnodes[i].common.public_flag = TREE_VIA_PUBLIC(base_binfo);
	    tmpnodes[i].common.protected_flag = TREE_VIA_PROTECTED(base_binfo);
	    if (!(TREE_VIA_PUBLIC(base_binfo) || TREE_VIA_PROTECTED(base_binfo))){
	      tmpnodes[i].common.private_flag = 1;
	    } else {
	      tmpnodes[i].common.private_flag = 0;
	    }
	    /* use unsigned flag for virtual inheritance */
	    tmpnodes[i].common.unsigned_flag = TREE_VIA_VIRTUAL(base_binfo);
	  }
	  the_graph.nodeBuffer[the_graph.curNodePos].record_type.binf = 
	    queue_and_return_index(di,tmpnodes,DUMP_NONE)-1;
	} else {
	  the_graph.nodeBuffer[the_graph.curNodePos].record_type.binf = -1;
	}
      }
      return 1;
      break;
    case UNION_TYPE:
      if (TYPE_PTRMEMFUNC_P(t)){
	the_graph.nodeBuffer[the_graph.curNodePos].union_type.ptrmem = 1;
	the_graph.nodeBuffer[the_graph.curNodePos].union_type.ptd =
	  queue_and_return_index (di,TYPE_PTRMEM_POINTED_TO_TYPE(t),DUMP_NONE)-1;
	the_graph.nodeBuffer[the_graph.curNodePos].union_type.cls =
	  queue_and_return_index (di,TYPE_PTRMEM_CLASS_TYPE(t),DUMP_NONE)-1;
	the_graph.nodeBuffer[the_graph.curNodePos].union_type.vfld = -1;
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].union_type.ptrmem = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].union_type.ptd = -1;
	the_graph.nodeBuffer[the_graph.curNodePos].union_type.cls = -1;
	the_graph.nodeBuffer[the_graph.curNodePos].union_type.vfld =
	  queue_and_return_index (di,TYPE_VFIELD(t),DUMP_NONE)-1;
      }
      /* gcc uses variable size nodes here. GD does not have variable
      * size nodes, so we transform the vector into a list */
      {
	/* number of base classes */
	int n_baseclasses = CLASSTYPE_N_BASECLASSES(t);
	if (n_baseclasses != 0) {
	  tree binfos = TYPE_BINFO_BASETYPES(t);
	  tree tmpnodes;
	  int i;
	  tmpnodes = malloc((unsigned)(sizeof(union tree_node) * n_baseclasses));
	  if (tmpnodes == NULL) {
	    fprintf(stderr,"could not allocate extra nodes for inheritance\n");
	    exit(1);
	  }
	  /* fill array */
	  for(i = 0; i < n_baseclasses; i++) {
	    tree base_binfo = TREE_VEC_ELT(binfos,i);
	    tmpnodes[i].common.code=TREE_LIST;
	    tmpnodes[i].common.type=NULL;
	    if (i != n_baseclasses -1){
	      tmpnodes[i].common.chain=(tree)&(tmpnodes[i+1]);
	    } else{
	      tmpnodes[i].common.chain=NULL;
	    }
	    tmpnodes[i].list.purpose=NULL;
	    tmpnodes[i].list.value= BINFO_TYPE(base_binfo);
	    tmpnodes[i].common.public_flag = TREE_VIA_PUBLIC(base_binfo);
	    tmpnodes[i].common.protected_flag = TREE_VIA_PROTECTED(base_binfo);
	    if (!(TREE_VIA_PUBLIC(base_binfo) || TREE_VIA_PROTECTED(base_binfo))){
	      tmpnodes[i].common.private_flag = 1;
	    } else {
	      tmpnodes[i].common.private_flag = 0;
	    }
	    /* use unsigned flag for virtual inheritance */
	    tmpnodes[i].common.unsigned_flag = TREE_VIA_VIRTUAL(base_binfo);
	  }
	  the_graph.nodeBuffer[the_graph.curNodePos].union_type.binf = 
	    queue_and_return_index(di,tmpnodes,DUMP_NONE)-1;
	} else {
	  the_graph.nodeBuffer[the_graph.curNodePos].union_type.binf = -1;
	}
      }
      return 1;
      break;
    case QUAL_UNION_TYPE:
    case LANG_TYPE:
      break;
    case BOUND_TEMPLATE_TEMPLATE_PARM:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBOUND_TEMPLATE_TEMPLATE_PARM;
      break;
    case TEMPLATE_TYPE_PARM:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTEMPLATE_TYPE_PARM;
      break;
    case TEMPLATE_PARM_INDEX:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTEMPLATE_PARM_INDEX;
      break;
    case TEMPLATE_TEMPLATE_PARM:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTEMPLATE_TEMPLATE_PARM;
      break;
    case TYPENAME_TYPE:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTYPENAME_TYPE;
      break;
    case TYPEOF_TYPE:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTYPEOF_TYPE;
      break;
    case COMPOUND_STMT:
    case EXPR_STMT:
    case DECL_STMT:
    case SCOPE_STMT:
    case RETURN_STMT:
    case STMT_EXPR:
    case BREAK_STMT:
    case CONTINUE_STMT:
    case IF_STMT:
    case SWITCH_STMT:
    case LABEL_STMT:
      break;
    case TRY_BLOCK:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRY_BLOCK;
      the_graph.nodeBuffer[the_graph.curNodePos].try_block.line = 
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].try_block.cleanup =
	(CLEANUP_P(t))?1:0;
      the_graph.nodeBuffer[the_graph.curNodePos].try_block.body =
	  queue_and_return_index(di,TRY_STMTS(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].try_block.hdlr =
	  queue_and_return_index(di,TRY_HANDLERS(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].try_block.next =
	  queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    case HANDLER:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gHANDLER;
      the_graph.nodeBuffer[the_graph.curNodePos].handler.line = 
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].handler.body =
	  queue_and_return_index(di,HANDLER_BODY(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].handler.next =
	  queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    case START_CATCH_STMT:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSTART_CATCH_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].start_catch_stmt.line = 
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].start_catch_stmt.catch_type =
	  queue_and_return_index(di,TREE_TYPE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].start_catch_stmt.next =
	  queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    case CLEANUP_STMT:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCLEANUP_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].cleanup_stmt.line = 
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].cleanup_stmt.decl =
	  queue_and_return_index(di,CLEANUP_DECL(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].cleanup_stmt.clean_expr =
	  queue_and_return_index(di,CLEANUP_EXPR(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].cleanup_stmt.next =
	  queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    case CASE_LABEL:
    case GOTO_STMT:
    case FOR_STMT:
    case WHILE_STMT:
    case DO_STMT:
    case ASM_STMT:
      break;
    /* ref */
    case COMPONENT_REF:
    case BIT_FIELD_REF:
    case INDIRECT_REF:
    case BUFFER_REF:
    case ARRAY_REF:
    case OFFSET_REF:
      break;
    case SCOPE_REF:
/* ??? */
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSCOPE_REF;
      break;
    case MEMBER_REF:
      break;
    case CONSTRUCTOR:
    case COMPOUND_EXPR:
    case MODIFY_EXPR:
    case INIT_EXPR:
      break;
    case THROW_EXPR:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTHROW_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].throw_expr.op0 =
	  queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      break;
    case TARGET_EXPR:
    case COND_EXPR:
    case BIND_EXPR:
    case CALL_EXPR:
    case METHOD_CALL_EXPR:
    case WITH_CLEANUP_EXPR:
    case CLEANUP_POINT_EXPR:
    case PLACEHOLDER_EXPR:
    case WITH_RECORD_EXPR:
    case PLUS_EXPR:
    case MINUS_EXPR:
    case MULT_EXPR:
    case TRUNC_DIV_EXPR:
    case CEIL_DIV_EXPR:
    case FLOOR_DIV_EXPR:
    case ROUND_DIV_EXPR:
    case TRUNC_MOD_EXPR:
    case CEIL_MOD_EXPR:
    case FLOOR_MOD_EXPR:
    case ROUND_MOD_EXPR:
    case RDIV_EXPR:
    case EXACT_DIV_EXPR:
    case FIX_TRUNC_EXPR:
    case FIX_CEIL_EXPR:
    case FIX_FLOOR_EXPR:
    case FIX_ROUND_EXPR:
    case FLOAT_EXPR:
    case EXPON_EXPR:
    case NEGATE_EXPR:
    case MIN_EXPR:
    case MAX_EXPR:
    case ABS_EXPR:
    case FFS_EXPR:
    case LSHIFT_EXPR:
    case RSHIFT_EXPR:
    case LROTATE_EXPR:
    case RROTATE_EXPR:
    case BIT_IOR_EXPR:
    case BIT_XOR_EXPR:
    case BIT_AND_EXPR:
    case BIT_ANDTC_EXPR:
    case BIT_NOT_EXPR:
    case TRUTH_ANDIF_EXPR:
    case TRUTH_ORIF_EXPR:
    case TRUTH_AND_EXPR:
    case TRUTH_OR_EXPR:
    case TRUTH_XOR_EXPR:
    case TRUTH_NOT_EXPR:
    case LT_EXPR:
    case LE_EXPR:
    case GT_EXPR:
    case GE_EXPR:
    case EQ_EXPR:
    case NE_EXPR:
    case UNORDERED_EXPR:
    case ORDERED_EXPR:
    case UNLT_EXPR:
    case UNLE_EXPR:
    case UNGT_EXPR:
    case UNGE_EXPR:
    case UNEQ_EXPR:
    case IN_EXPR:
    case SET_LE_EXPR:
    case CARD_EXPR:
    case RANGE_EXPR:
    case CONVERT_EXPR:
    case NOP_EXPR:
    case NON_LVALUE_EXPR:
    case SAVE_EXPR:
    case UNSAVE_EXPR:
    case RTL_EXPR:
    case ADDR_EXPR:
    case REFERENCE_EXPR:
    case ENTRY_VALUE_EXPR:
    case COMPLEX_EXPR:
    case CONJ_EXPR:
    case REALPART_EXPR:
    case PREDECREMENT_EXPR:
    case PREINCREMENT_EXPR:
    case POSTDECREMENT_EXPR:
    case POSTINCREMENT_EXPR:
    case VA_ARG_EXPR:
    case TRY_CATCH_EXPR:
    case TRY_FINALLY_EXPR:
    case POPDHC_EXPR:
    case POPDCC_EXPR:
    case LABEL_EXPR:
    case GOTO_EXPR:
    case LOOP_EXPR:
    case LABELED_BLOCK_EXPR:
    case EXIT_BLOCK_EXPR:
    case SWITCH_EXPR:
    case SIZEOF_EXPR:
    case ARROW_EXPR:
    case ALIGNOF_EXPR:
    case TYPE_EXPR:
      break;
    case AGGR_INIT_EXPR:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gAGGR_INIT_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].aggr_init_expr.ctor =
	AGGR_INIT_VIA_CTOR_P(t);
      the_graph.nodeBuffer[the_graph.curNodePos].aggr_init_expr.fn =
	  queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].aggr_init_expr.args =
	  queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].aggr_init_expr.decl =
	  queue_and_return_index(di,TREE_OPERAND(t,2),DUMP_NONE)-1;
      break;
    case EMPTY_CLASS_EXPR:
      break;
    case TEMPLATE_ID_EXPR:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTEMPLATE_ID_EXPR;
      break;
    case LOOKUP_EXPR:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLOOKUP_EXPR;
      break;
    case FUNCTION_NAME:
      break;
    case SUBOBJECT:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSUBOBJECT;
      the_graph.nodeBuffer[the_graph.curNodePos].subobject.line = 
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].subobject.clnp = 
	  queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].subobject.next = 
	  queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    case CTOR_STMT:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCTOR_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].ctor_stmt.line = 
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].ctor_stmt.beg_end = 
	(CTOR_BEGIN_P(t))?1:0;
      the_graph.nodeBuffer[the_graph.curNodePos].ctor_stmt.next = 
	  queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    case CTOR_INITIALIZER:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCTOR_INITIALIZER;
      break;
    case RETURN_INIT:
    case TAG_DEFN:
    case IDENTITY_CONV:
    case LVALUE_CONV:
    case QUAL_CONV:
    case STD_CONV:
    case PTR_CONV:
    case PMEM_CONV:
    case BASE_CONV:
    case REF_BIND:
    case AMBIG_CONV:
    case RVALUE_CONV:
    case INTEGER_CST:
    case STRING_CST:
    case COMPLEX_CST:
    case REAL_CST:
      break;
    case PTRMEM_CST:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPTRMEM_CST;
      the_graph.nodeBuffer[the_graph.curNodePos].ptrmem_cst.clas =
	  queue_and_return_index(di,PTRMEM_CST_CLASS(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].ptrmem_cst.mbr =
	  queue_and_return_index(di,PTRMEM_CST_MEMBER(t),DUMP_NONE)-1;
      break;
    case OP_IDENTIFIER:
      break;
    case IDENTIFIER_NODE:
      if (IDENTIFIER_OPNAME_P (t)) {
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.op_type_un=1;
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.tynm=-1;
	return 1;
      } else if (IDENTIFIER_TYPENAME_P (t)) {
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.op_type_un=2;
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.tynm=
	  queue_and_return_index(di,TREE_TYPE(t),DUMP_NONE)-1;
	return 1;
      } else if (t == anonymous_namespace_name) {
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.op_type_un=3;
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.tynm=-1;
	return 1;
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.op_type_un=0;
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.tynm=-1;
      }
      break;
    case TREE_LIST:
    case TREE_VEC:
    case ERROR_MARK:
    case DEFAULT_ARG:
      break;
    case OVERLOAD:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gOVERLOAD;
      the_graph.nodeBuffer[the_graph.curNodePos].overload.crnt =
	  queue_and_return_index(di,OVL_CURRENT(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].overload.chan =
	  queue_and_return_index(di,OVL_CHAIN(t),DUMP_NONE)-1;
      break;
/* new???*/
    case REINTERPRET_CAST_EXPR:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gREINTERPRET_CAST_EXPR;
      break;
    case MODOP_EXPR:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gMODOP_EXPR;
      break;
    case STATIC_CAST_EXPR:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSTATIC_CAST_EXPR;
      break;
    case CAST_EXPR:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCAST_EXPR;
      break;
    case CONST_CAST_EXPR:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCONST_CAST_EXPR;
      break;
    default:
      fprintf(stderr,"unkown tree code type %d in cp/dump.c\n",code);
      exit(1);
  }
  return 0;

}
