/*+
 * File:	gd_graph.h
 *
 * Purpose:	This file provides the definitions for the
 * 		gd_graph used to convert GCC Abstract Syntax
 * 		Graphs(ASG) to Datrix ASGs in GXL. It merges
 * 		both types of nodes, as well as some intermediate
 * 		nodes.
 *
 * Author:	Thomas R. Dean
 *
 * Revision History:
 *
 *  1.0 - January-February 2001
 *      - initial version
 *
-*/

#ifndef _GD_GRAPH_H_
#define _GD_GRAPH_H_

/* version control for binary gd_graph files */
#define GRAPH_MAJOR 3
#define GRAPH_MINOR 0

#define GD_NODE_SIZE 64

/***************************************************
 * Graph Node Definitions
 ***************************************************/

/* Graph Node types - union of the types from GCC
 *     Datrix and some internal types used during
 *     the conversion                                */

enum gd_node_type {
  /* node types from GCC */
#define DEFTREECODE(SYM,STR,A,B) g ## SYM,
#include "tree.def"
#include "c-common.def"
#include "cp/cp-tree.def"
#undef DEFTREECODE

  /* end of node types from GCC */

  /* node types from DATRIX */
  dC_ASG_ND,
    dC_IDENTIFIER,
      dC_TYPE,
	dC_BUILT_IN_TYPE,
	dC_AGGR_TYPE,
	dC_ENUM_TYPE,
	dC_TEMPL_PARM_TYPE,
	dC_GENERIC_TYPE,
	dC_GENERATED_TYPE,
	dC_ALIAS_TYPE,
	dC_FORWARD_TYPE,
	dC_ARRAY_TYPE,
	dC_PTR_TYPE,
	dC_REF_TYPE,
	dC_FCT_TYPE,
      dC_SCOPE,
	dC_SYSTEM,
	dC_SCOPE_GLB,
	dC_SCOPE_COMPIL,
	dC_SCOPE_FILE,
	dC_PACKAGE,
	dC_NAME_SPACE,
	dC_BLOCK,
	dC_STATIC_BLOCK,
	dC_TRY_BLOCK,
	dC_CATCH_BLOCK,
	dC_FINALLY_BLOCK,
	dC_SYNCHRONIZED_BLOCK,
      dC_OBJECT,
	dC_FORMAL_FCT_PARAM,
	dC_FORMAL_CATCH_PARAM,
      dC_FUNCTION,
      dC_LABEL,
	dC_CASE_LABEL,
	dC_DEFAULT_LABEL,
      dC_ENUMERATOR,
    dC_EXPRESSION,
      dC_EMPTY_EXPR,
      dC_LITERAL,
      dC_SYNTAX_ERROR,
      dC_NAME_REF,
      dC_OPERATOR,
	dC_UNARY_OPERATOR,
	dC_BINARY_OPERATOR,
	dC_QUESTION_OPERATOR,
      dC_INIT_LIST,
      dC_ACTUAL_FCT_PARAM_LIST,
      dC_ACTUAL_TEMPL_PARAM_LIST,
      dC_CREATION_EXPR,
      dC_DESTROY_EXPR,
    dC_CNTRL_STMT,
      dC_SELECTION,
	dC_IF,
	dC_SWITCH,
      dC_JUMP,
	dC_GOTO,
	dC_RETURN,
	dC_BREAK,
	dC_CONTINUE,
      dC_LOOP,
	dC_DO_WHILE_LOOP,
	dC_WHILE_DO_LOOP,
	dC_FOR_LOOP,
    dC_SQL_STMT,
    dC_COMMENT,
      dC_JAVA_DOC_COMMENT,
    dC_THROW_SPEC,
    dC_ASM_BLOCK,

  dC_ARC,
    dC_ARC_REFERENCE,
      dC_INSTANCE,
      dC_INHERITANCE,
      dC_IMPLEMENTS,
      dC_ARC_FRIEND,
      dC_ARC_DEF_LOC,
      dC_REFERS_TO,
    dC_ARC_SYNTAX,
      dC_ARC_SON,
      dC_ARC_TRUE,
      dC_ARC_FALSE,
      dC_ARC_OPD,
      dC_IMPORT,
      dC_ARC_COND,
      dC_ARC_FOR_INIT,
      dC_ARC_FOR_INCR,
      dC_ARC_INIT_VAL,
      dC_ARC_BIT_DIM,
      dC_ARC_ARRAY_DIM,

  /* end of types from DATRIX */

  /* deleted node */
  NODE_DELTED,

  /* internal use nodes types */

  NODE_NONE,
  /* end internal use node types */
  NUM_NODES
};

extern const char *gd_node_type_ascii[];

/*****************************
 * index types
 *****************************/

/* the compiler doesn't actually enforce the differences
 * between these, but it will help for documentation */

typedef int str_ptr; /* index into string table */
typedef int node_ptr; /* index into node table */

/* other types */

typedef enum {
  src_lang_none = 0,
  src_lang_c = 1,
  src_lang_cp,
  src_lang_java
} src_lang;

/* Graph Node definitions */

/* we use a similar strategy to that used by gcc,
 * we have a common header for each node. It is included
 * by each node structure (to keep space alignment correct)
 * and is is also included in the union for easy access.
 *
 * only includes the fields we are interested in
 *
 * */

typedef struct {
  enum gd_node_type type:16;
  /* bit fields for simple definitions */
  unsigned unsigned_flag: 1; /* 17 */
  unsigned constant_flag: 1; /* 18 */
  unsigned static_flag:   1; /* 19 */
  unsigned public_flag:   1; /* 20 */
  unsigned private_flag:  1; /* 21 */
  unsigned protected_flag:1; /* 22 */ /* 10 bits left */
  /* for use of conversion algorithm */
  unsigned mark_flag1:    1; /* 23 */ /* 9 bits left */
  unsigned mark_flag2:    1; /* 24 */ /* 8 bits left */
  node_ptr other_node;
} gd_graph_common;

/* names of subfields of these structures comes from
 * the dump file created by c-dump.c */
typedef struct {
  gd_graph_common common;
  node_ptr	name;
  node_ptr	mngl;
  node_ptr	type;
  node_ptr	chan;
  node_ptr	scpe;
  str_ptr	srcf;
  int 		srcl;
  src_lang	lang;
  int		artificial;
} t_gcc_decl;

typedef struct {
  t_gcc_decl	decl;
  node_ptr	dcls;	/* C++ only */
  node_ptr	alis;	/* C++ only */
} t_namespace_decl;

/* function decl */
typedef struct {
  t_gcc_decl	decl;
  node_ptr args;
  node_ptr body;
  unsigned undef:1;      /* 0 = no, 1 = undefined */
  unsigned ext_static:1; /* 0 = static, 1 = extern */
  /* c++ stuff */
  unsigned member:1;
  unsigned constr:1;	/* constructor */
  unsigned decon:1;
  unsigned oper:1;	/* operator */
  unsigned conver:1;	/* conversion */
  unsigned ginit:1;	/* global init */
  unsigned gfini:1;	/* global fini */
} t_function_decl;

typedef struct {
  t_gcc_decl	decl;
  node_ptr	init;
  node_ptr	size;
  int		algn;
  int		used;
  int		reg;
} t_var_decl;

typedef struct {
  t_gcc_decl	decl;
  node_ptr	argt;
  node_ptr	size;
  int		algn;
  int		used;
  int		reg;
} t_parm_decl;

typedef struct {
  t_gcc_decl	decl;
} t_type_decl;

typedef struct {
  t_gcc_decl	decl;
} t_label_decl;

typedef struct {
  t_gcc_decl	decl;
  node_ptr	cnst;
} t_const_decl;

typedef struct {
  t_gcc_decl	decl;
  node_ptr	init;
  node_ptr	size;
  int		algn;
} t_result_decl;

typedef struct {
  t_gcc_decl	decl;
  node_ptr	init;
  node_ptr	size;
  int		algn;
  int		bitfield;
  node_ptr	bpos;
} t_field_decl;

typedef struct {
  t_gcc_decl	decl;
  node_ptr	rslt;
  node_ptr	inst;
  node_ptr	spcs;
  node_ptr	parms;
} t_template_decl;

typedef struct {
  t_gcc_decl	decl;
} t_using_decl;


typedef struct {
  gd_graph_common common;
  node_ptr	name;
  node_ptr	size;
  int		algn;
  node_ptr	unql;
  /* flags */
  unsigned	qual_const:1;
  unsigned	qual_volatile:1;
  unsigned	qual_restrict:1;
} t_gcc_type;

typedef struct {
  t_gcc_type type;
  node_ptr retn;
  node_ptr prms;
} t_function_type;

typedef struct {
  t_gcc_type	type;
  int		prec;
  int		unsgnd;
  node_ptr	min;
  node_ptr	max;
} t_integer_type;

typedef struct {
  t_gcc_type	type;
  node_ptr	elts;
  node_ptr	domn;
} t_array_type;

typedef struct {
  t_gcc_type	type;
} t_void_type;

typedef struct {
  t_gcc_type	type;
  int		prec;
} t_real_type;

typedef struct {
  t_gcc_type	type;
} t_complex_type;

typedef struct {
  t_gcc_type	type;
} t_vector_type;

typedef struct {
  /* same as t_integer_type but one more field, csts */
  t_gcc_type	type;
  int		prec;
  int		unsgnd;
  node_ptr	min;
  node_ptr	max;
  node_ptr	csts;
} t_enumeral_type;

typedef struct {
  t_gcc_type	type;
} t_boolean_type;

typedef struct {
  t_gcc_type	type;
} t_char_type;

typedef struct {
  t_gcc_type	type;
  node_ptr	ptd;
  node_ptr	cls;
  int		ptrmem; /* 0 no 1 yes */
} t_pointer_type;

typedef struct {
  t_gcc_type	type;
} t_offset_type;

typedef struct {
  t_gcc_type	type;
  node_ptr	refd;
} t_reference_type;

typedef struct {
  t_gcc_type	type;
  node_ptr	clas;
  node_ptr retn;
  node_ptr prms;
} t_method_type;

typedef struct {
  t_gcc_type	type;
} t_file_type;

typedef struct {
  t_gcc_type	type;
} t_set_type;

typedef struct {
  t_gcc_type	type;
  node_ptr	flds;
  node_ptr	fncs;
  int		ptrmem;
  int		isClass;
  node_ptr	ptd;
  node_ptr	cls;
  node_ptr	vfld;
  node_ptr	binf;
} t_record_type;

typedef struct {
  t_gcc_type	type;
  node_ptr	flds;
  node_ptr	fncs;
  int		ptrmem;
  node_ptr	ptd; /* only valid if ptrmem = 1*/
  node_ptr	cls; /* only valid if ptrmem = 1*/
  node_ptr	vfld;
  node_ptr	binf;
} t_union_type;

typedef struct {
  t_gcc_type	type;
} t_qual_union_type;

typedef struct {
  t_gcc_type	type;
} t_lang_type;

typedef struct {
  t_gcc_type	type;
} t_template_type_parm;

/* ??? */
typedef struct {
  gd_graph_common common; /* code class 'x' */
} t_template_parm_index;

typedef struct {
  t_gcc_type	type;
} t_template_template_parm;

typedef struct {
  t_gcc_type	type;
} t_bound_template_template_parm;

typedef struct {
  t_gcc_type	type;
} t_typename_type;

typedef struct {
  t_gcc_type	type;
} t_typeof_type;

/* expressions */

typedef struct {
  gd_graph_common common;
  node_ptr	type;
} t_gcc_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
} t_gcc_expr1;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_gcc_expr2;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_gcc_expr_ineq;

/* statements */
/* block is code class 'b' */
typedef struct {
  gd_graph_common common;
} t_block;

/* code class 'e' */

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	body;
  node_ptr	next;
} t_compound_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	texpr;
  node_ptr	next;
} t_expr_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	decl;
  node_ptr	next;
} t_decl_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  unsigned	begin_end:1; /* 0 = begin, 1= end */
  unsigned	null:1;
  unsigned	clnp:1;
  node_ptr	next;
} t_scope_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	ret_expr;
  node_ptr	next;
} t_return_stmt;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	stmt;
} t_stmt_expr;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	next;
} t_break_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	next;
} t_continue_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	cond;
  node_ptr	thenc;
  node_ptr	elsec;
  node_ptr	next;
} t_if_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	cond;
  node_ptr	body;
  node_ptr	next;
} t_switch_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	label;
  node_ptr	next;
} t_label_stmt;

typedef struct { /* C++ only */
  t_gcc_expr	expr;
  int		line;
  int		cleanup;
  node_ptr	body;
  node_ptr	hdlr;
  node_ptr	next;
} t_try_block;

typedef struct { /* C++ only */
  t_gcc_expr	expr;
  int		line;
  node_ptr	body;
  node_ptr	next;
} t_handler;

typedef struct { /* C++ only */
  t_gcc_expr	expr;
  int		line;
  node_ptr	catch_type; /* another type? */
  node_ptr	next;
} t_start_catch_stmt;

typedef struct { /* C++ only */
  t_gcc_expr	expr;
  int		line;
  node_ptr	decl;
  node_ptr	clean_expr;
  node_ptr	next;
} t_cleanup_stmt;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	low;
  node_ptr	high;
  node_ptr	next;
} t_case_label;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	dest;
  node_ptr	next;
} t_goto_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	init;
  node_ptr	cond;
  node_ptr	inc_expr;
  node_ptr	body;
  node_ptr	next;
} t_for_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	body;
  node_ptr	cond;
  node_ptr	next;
} t_while_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  node_ptr	body;
  node_ptr	cond;
  node_ptr	next;
} t_do_stmt;

typedef struct {
  t_gcc_expr	expr;
  int		line;
  int		vol;
  node_ptr	strg;
  node_ptr	outs;
  node_ptr	ins;
  node_ptr	clbr;
  node_ptr	next;
} t_asm_stmt;


/* refs */
typedef struct {
  gd_graph_common common;
  node_ptr	op0;
  node_ptr	op1;
} t_component_ref;

typedef struct {
  gd_graph_common common;
} t_bit_field_ref;

typedef struct {
  gd_graph_common common;
  node_ptr	op0;
} t_indirect_ref;

typedef struct {
  gd_graph_common common;
} t_buffer_ref;

typedef struct {
  gd_graph_common common;
  node_ptr	op0;
  node_ptr	op1;
} t_array_ref;

typedef struct {
  gd_graph_common common;
  node_ptr	op0;
  node_ptr	op1;
} t_offset_ref;

typedef struct {
  gd_graph_common common;
  node_ptr	op0;
  node_ptr	op1;
} t_scope_ref;

typedef struct {
  gd_graph_common common;
  node_ptr	op0;
  node_ptr	op1;
} t_member_ref;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	elts;
}t_constructor;

/* expressions */
typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_compound_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_modify_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_init_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
} t_throw_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	decl;
  node_ptr	init;
  node_ptr	clnp;
} t_target_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
  node_ptr	op2;
} t_cond_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	vars;
  node_ptr	body;
} t_bind_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	fn;
  node_ptr	args;
} t_call_expr;

typedef struct {
  t_gcc_expr	expr;
} t_method_call_expr;

typedef struct {
  t_gcc_expr	expr;
} t_with_cleanup_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
} t_cleanup_point_expr;

typedef struct { /* class_code 'x' */
  gd_graph_common common;
} t_placeholder_expr;

typedef struct {
  t_gcc_expr	expr;
} t_with_record_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_plus_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_minus_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_mult_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_trunc_div_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_ceil_div_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_floor_div_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_round_div_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_trunc_mod_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_ceil_mod_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_floor_mod_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_round_mod_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_rdiv_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_exact_div_expr;

typedef struct {
  t_gcc_expr1	expr1;
} t_fix_trunc_expr;

typedef struct {
  t_gcc_expr1	expr1;
} t_fix_ceil_expr;

typedef struct {
  t_gcc_expr1	expr1;
} t_fix_floor_expr;

typedef struct {
  t_gcc_expr1	expr1;
} t_fix_round_expr;

typedef struct {
  t_gcc_expr1	expr1;
} t_float_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_expon_expr;

typedef struct {
  t_gcc_expr1	expr1;
} t_negate_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_min_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_max_expr;

typedef struct {
  t_gcc_expr1	expr1;
} t_abs_expr;

typedef struct {
  t_gcc_expr1	expr1;
} t_ffs_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_lshift_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_rshift_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_lrotate_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_rrotate_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_bit_ior_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_bit_xor_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_bit_and_expr;

typedef struct {
  t_gcc_expr2	expr2;
} t_bit_andtc_expr;

typedef struct {
  t_gcc_expr1	expr1;
} t_bit_not_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
  node_ptr	op1;
} t_truth_andif_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
  node_ptr	op1;
} t_truth_orif_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
  node_ptr	op1;
} t_truth_and_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
  node_ptr	op1;
} t_truth_or_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
  node_ptr	op1;
} t_truth_xor_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_truth_not_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_lt_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_le_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_gt_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_ge_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_eq_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_ne_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_unordered_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_ordered_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_unlt_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_unle_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_ungt_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_unge_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_uneq_expr;

typedef struct {
  t_gcc_expr2 expr2;
} t_in_expr;

typedef struct {
  t_gcc_expr_ineq expr_ineq;
} t_set_le_expr;

typedef struct {
  t_gcc_expr1 expr1;
} t_card_expr;

typedef struct {
  t_gcc_expr2 expr2;
} t_range_expr;

typedef struct {
  t_gcc_expr1 expr1;
} t_convert_expr;

typedef struct {
  t_gcc_expr1 expr1;
} t_nop_expr;

typedef struct {
  t_gcc_expr1 expr1;
} t_non_lvalue_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_save_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
} t_unsave_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
} t_rtl_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
} t_addr_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
} t_reference_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
} t_entry_value_expr;

typedef struct {
  t_gcc_expr2 expr2;
} t_complex_expr;

typedef struct {
  t_gcc_expr1 expr1;
} t_conj_expr;

typedef struct {
  t_gcc_expr1 expr1;
} t_realpart_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_predecrement_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_preincrement_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_postdecrement_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_postincrement_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
} t_va_arg_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_try_catch_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_try_finally_expr;

typedef struct {
  t_gcc_expr	expr;
  node_ptr	op0;
  node_ptr	op1;
} t_goto_subroutine_expr;

/* code class s */
typedef struct {
  gd_graph_common common;
} t_popdhc_expr;

typedef struct {
  gd_graph_common common;
} t_popdcc_expr;

typedef struct {
  gd_graph_common common;
} t_label_expr;

typedef struct {
  gd_graph_common common;
} t_goto_expr;

typedef struct {
  gd_graph_common common;
} t_return_expr;

typedef struct {
  gd_graph_common common;
  node_ptr	cond;
} t_exit_expr;

typedef struct {
  gd_graph_common common;
  node_ptr	body;
} t_loop_expr;

/* code class 'e' */
typedef struct {
  t_gcc_expr	expr; /* code class e */
} t_labeled_block_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
} t_exit_block_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	expr2;
} t_expr_with_file_location;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
  node_ptr	op1;
} t_switch_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_sizeof_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_arrow_expr;
  
typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_alignof_expr;
  
typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_new_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_vec_new_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_delete_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_vec_delete_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_type_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  int		ctor;
  node_ptr	fn;
  node_ptr	args;
  node_ptr	decl;
} t_aggr_init_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
} t_empty_class_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
  node_ptr	op1;
} t_template_id_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_lookup_expr;

typedef struct {
  t_gcc_expr	expr; /* code class e */
} t_function_name;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  int		line;
  node_ptr	clnp;
  node_ptr	next;
} t_subobject;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  int		line;
  int		beg_end;
  node_ptr	next;
} t_ctor_stmt;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
  node_ptr	op1;
} t_ctor_initializer;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
  node_ptr	op1;
} t_return_init;

typedef struct {
  t_gcc_expr	expr; /* code class e */
} t_tag_defn;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_identity_conv;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_lvalue_conv;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_qual_conv;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_std_conv;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_ptr_conv;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_pmem_conv;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_base_conv;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_ref_bind;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
  node_ptr	op1;
} t_user_conv;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_ambig_conv;

typedef struct {
  t_gcc_expr	expr; /* code class e */
  node_ptr	op0;
} t_rvalue_conv;

typedef struct {
  gd_graph_common common;
} t_op_identifier;

typedef struct {
  gd_graph_common common;
  str_ptr strg;
  unsigned int lngt;
  int op_type_un; /* 0 empty, 1 operator, 2 typename, 3 unnamed */
  int	tynm; /* onlye defined if op_type_un = 2 */
} t_identifier_node;

typedef struct {
  gd_graph_common common;
  node_ptr	type;
} t_gcc_const;

typedef struct {
  t_gcc_const	gcc_cst;
  int		high; /* high word  -- if present */
  int		low; /* low word */
} t_integer_cst;

typedef struct {
  t_gcc_const gcc_cst;
  str_ptr value;
} t_real_cst;

typedef struct {
  t_gcc_const gcc_cst;
  node_ptr	clas;
  node_ptr	mbr;
} t_ptrmem_cst;

typedef struct {
  t_gcc_const gcc_cst;
} t_complex_cst;

typedef struct {
  t_gcc_const	gcc_cst;
  str_ptr	strg;
  int		lngt;
} t_string_cst;

typedef struct {
  gd_graph_common common;
  node_ptr	purpose;
  node_ptr	value;
  node_ptr	chan;
} t_tree_list;

typedef struct {
  gd_graph_common common;
  int binfo;
  node_ptr inherit;
  node_ptr type;
  node_ptr base;
  unsigned pub:1;
  unsigned prot:1;
  unsigned priv:1;
  unsigned virt:1;
} t_tree_vec;

typedef struct {
  gd_graph_common common;
} t_error_mark;

typedef struct {
  gd_graph_common common;
} t_default_arg;

typedef struct {
  gd_graph_common common;
} t_cplus_binding;

typedef struct {
  gd_graph_common common;
  node_ptr	crnt;
  node_ptr	chan;
} t_overload;

typedef struct {
  gd_graph_common common;
} t_wrapper;

/***************
 * Datrix Types
 ***************/

typedef struct {
  gd_graph_common common;
  str_ptr	file;
  int		line;			/* datrix wants begin and end: line.col
					   gcc only gives us lines, and only 
					   for statements*/
} t_dc_asg_nd;

typedef enum {
  visb_unk = 0, visb_pri, visb_pro, visb_pub
} t_visb;

typedef struct  {
  t_dc_asg_nd	dc_asg_nd;
  str_ptr	name;
  t_visb	visb;
  node_ptr	instance;  /* arc Instance, other end is _type */
  node_ptr	defLoc;    /* ArcDefLoc other end is cIdentifier */
  node_ptr	arcFriend; /* ArcFriend other end is cType */
  node_ptr	array_dim; /* cArcArrayDim other end is cExpression */
} t_dc_identifier;

typedef struct {
  t_dc_identifier dc_identifier;
   node_ptr	implements; /* arc Implements, other end is type */
} t_dc_type;

typedef struct {
  t_dc_type	type;
} t_dc_built_in_type;

typedef enum {
  aggr_class, aggr_record, aggr_struct, aggr_union,
  aggr_area, aggr_interface
} t_aggr_enum;

typedef struct {
  t_dc_type	type;
  t_aggr_enum	aggr_type;
  node_ptr	members;	/* points to gd_node of cArcSon, other end of
				   cArcson is cIdentifier */
  node_ptr	inheritance;	/* points to cInheritance, other end is cType */
} t_dc_aggr_type;

typedef struct {
  t_dc_type	type;
  node_ptr	values;	/* gives cArcSon, other end is cEnumerator */
} t_dc_enum_type;

typedef struct {
  t_dc_type	type;
} t_dc_templ_parm_type;

typedef struct {
  t_dc_type	type;
  node_ptr	data_functs; /* cArcSon, other end is cIdentifier */
  node_ptr	formals;     /* cArcSon, other end is cTemplParmType */
} t_dc_generic_type;

typedef struct {
  t_dc_type	type;
  node_ptr	actuals;     /* cArcSon, other end is cExpression */
} t_dc_generated_type;

typedef struct {
  t_dc_type	type;
  node_ptr	original;    /* cInstance, other end is original type */
} t_dc_alias_type;

typedef struct {
  t_dc_type	type;
} t_dc_forward_type;

typedef struct {
  t_dc_type	type;
} t_dc_array_type;

typedef struct {
  t_dc_type	type;
} t_dc_ptr_type;

typedef struct {
  t_dc_type	type;
} t_dc_ref_type;

typedef struct {
  t_dc_type	type;
  node_ptr	formalParams;	/* gives cArcSon, other end is cFormalFctParam */
} t_dc_fct_type;

typedef struct {
  t_dc_identifier dc_identifier;
  node_ptr	contents;	/* cArcSon other end is cAsgNds */
} t_dc_scope;

typedef struct {
  t_dc_scope	scope;
} t_dc_system;

typedef struct {
  t_dc_scope	scope;
} t_dc_scope_glb;

typedef struct {
  t_dc_scope	scope;
} t_dc_scope_compile;

typedef struct {
  t_dc_scope	scope;
} t_dc_scope_file;

typedef struct {
  t_dc_scope	scope;
} t_dc_package;

typedef struct {
  t_dc_scope	scope;
} t_dc_name_space;

typedef struct {
  t_dc_scope	scope;
} t_dc_block;

typedef struct {
  t_dc_scope	scope;
} t_dc_static_block;

typedef struct {
  t_dc_scope	scope;
} t_dc_try_block;

typedef struct {
  t_dc_scope	scope;
} t_dc_catch_block;

typedef struct {
  t_dc_scope	scope;
} t_dc_finally_block;

typedef struct {
  t_dc_scope	scope;
} t_dc_synchronized_block;

typedef struct {
  t_dc_identifier dc_identifier;
  unsigned	static_flag:1;
  unsigned	extern_flag:1;
  unsigned	const_flag: 1;
  node_ptr	bitfield_dim;	/* cArcBitDim other end is cExpression */
  node_ptr	initial_val;	/* cArcInit_Val other end is cExpression */
} t_dc_object;

typedef struct {
  t_dc_object	object;
} t_dc_formal_fct_param;

typedef struct {
  t_dc_object	object;
} t_dc_formal_catch_param;

typedef struct {
  t_dc_identifier dc_identifier;
  unsigned	static_flag:1;
  unsigned	extern_flag:1;
  unsigned	const_flag:1;
  unsigned	virtual_code:2; /* 00 = non-virtual, 01 = virtual,
				   10 = pure-virtual */
  node_ptr	formal_fct;	/* cArcSon other end is cFormalFctParam */
  node_ptr	exception;	/* cArcSon other end is cThrowSpec */
  node_ptr	block;		/* cArcSon other end is cBlock */
} t_dc_function;

typedef struct {
  t_dc_identifier dc_identifier;
} t_dc_label;

typedef struct {
  t_dc_label	label;
  node_ptr	value;		/* cArcSon other end is cExpression */
} t_dc_case_label;

typedef struct {
  t_dc_label	label;
} t_dc_default_label;

typedef struct {
  t_dc_identifier dc_identifier;
  node_ptr	value;	/* gives cArcInitialVal, other end is cExpression */
} t_dc_enumerator;

typedef struct {
  t_dc_asg_nd	dc_asg_nd;
  node_ptr	refers; /* Arc cRefersTo opther end is cAsgNds */
} t_dc_expression;

typedef struct {
  t_dc_expression expression;
} t_dc_empty_expr;

typedef struct {
  t_dc_expression expression;
  str_ptr	value;	/* value as a string */
  node_ptr	type;	/* Datrix has an atrirbute here, we use an edge */
} t_dc_literal;

typedef struct {
  t_dc_expression expression;
} t_dc_syntax_error;

typedef struct {
  t_dc_expression expression;
  str_ptr	  name;
} t_dc_name_ref;

typedef enum {
  dc_uplus, dc_uminus, dc_pre_inc, dc_pre_dec, dc_post_inc, dc_post_dec,
  dc_bplus, dc_bminus, dc_modulo, dc_mult, dc_div, dc_bit_and, dc_bit_or,
  dc_bit_xor, dc_rshift, dc_lshift, dc_bit_not, dc_tst_eq, dc_tst_ne,
  dc_tst_ge, dc_tst_le, dc_tst_gt, dc_tst_lt, dc_tst_min, dc_tst_max,
  dc_not, dc_xor, dc_and, dc_or, dc_asgn_eq, dc_asgn_plus, dc_asgn_minus,
  dc_asgn_mult, dc_asgn_div, dc_asgn_mod, dc_asgn_lshift, dc_asgn_rshift,
  dc_asgn_bit_and, dc_asgn_bit_or, dc_asgn_bit_xor, dc_asgn_lt, dc_asgn_gt,
  dc_question, dc_addr, dc_deref, dc_pointsat, dc_pointsat_deref, dc_array_ref,
  dc_fct_call, dc_cast_call, dc_field_access, dc_field_deref, dc_sizeof,
  dc_comma, dc_const_cast, dc_static_cast, dc_reint_cast, dc_dynam_cast,
  dc_type_id, dc_sign_of, dc_range, dc_unk
} op_kind;

typedef struct {
  t_dc_expression expression;
  op_kind	op;
  node_ptr	ops;	/* cArcOpd/cArcOpdref other end is cExpression */
} t_dc_operator;

typedef struct {
  t_dc_operator optor;
} t_dc_unary_operator;

typedef struct {
  t_dc_operator optor;
} t_dc_binary_operator;

typedef struct {
  t_dc_operator optor;
} t_dc_question_operator;

typedef struct {
  t_dc_expression expression;
  node_ptr	init;	/* cArcSon, other end is cExpression */
} t_dc_init_list;

typedef struct {
  t_dc_expression expression;
  node_ptr	actuals; /* cArcSon, other end is cExpression */
} t_dc_actual_fct_param_list;

typedef struct {
  t_dc_expression expression;
  node_ptr	actuals; /* cArcSon, other end is cType */
} t_dc_actual_templ_param_list;

typedef struct {
  t_dc_expression expression;
  node_ptr	init;	/* cArcInitVal, other end is cExpression */
} t_dc_creation_expr;

typedef struct {
  t_dc_expression expression;
  node_ptr	what;	/* cArcSon, other end is cExpression */
  node_ptr	array;	/* cArcArrayDim, other end is cEmptyExpression */
} t_dc_destroy_expr;

typedef struct {
  t_dc_asg_nd	dc_asg_nd;
} t_dc_ctrl_stmt;

typedef struct {
  t_dc_ctrl_stmt	ctrl_stmt;
} t_dc_selection;

typedef struct {
  t_dc_ctrl_stmt	ctrl_stmt;
  node_ptr		cond;	    /* cArcCond - other end is cExpression */
  node_ptr		true_block; /* cArcTrue - other end is cBlock */
  node_ptr		false_block;/* cArcFalse - other end is cBlock */
} t_dc_if;

typedef struct {
  t_dc_ctrl_stmt	ctrl_stmt;
  node_ptr		cond; /* cArcCond - other end is cExpression */
  node_ptr		body; /* cArcFalse - other end is cBlock */
} t_dc_switch;

typedef struct {
  t_dc_ctrl_stmt	ctrl_stmt;
} t_dc_jump;

typedef struct {
  t_dc_jump 	jump;
  node_ptr	dest;	/* cArcSon other end is Expression */
} t_dc_goto;

typedef struct {
  t_dc_jump 	jump;
  node_ptr	value;	/* cArcSon other end is Expression */
} t_dc_return;

typedef struct {
  t_dc_jump 	jump;
  node_ptr	dest;	/* cArcSon other end is Expression -- Java only */
} t_dc_break;

typedef struct {
  t_dc_jump 	jump;
  node_ptr	dest;	/* cArcSon other end is Expression -- Java only */
} t_dc_continue;

typedef struct {
  t_dc_ctrl_stmt	ctrl_stmt;
} t_dc_loop;

typedef struct {
  t_dc_loop		loop;
  node_ptr		cond;	/* cArcCond - other end is cExpression */
  node_ptr		block;	/* cArcSon other end is cBlock */
} t_dc_do_while_loop;

typedef struct {
  t_dc_loop		loop;
  node_ptr		cond;	/* cArcCond - other end is cExpression */
  node_ptr		block;	/* cArcSon other end is cBlock */
} t_dc_while_do_loop;

typedef struct {
  t_dc_loop		loop;
  node_ptr		init;	/* cArcForInit other end is cAsgNd */
  node_ptr		cond;	/* cArcCond other end is cExpression */
  node_ptr		incr;	/* cArcForIncr other end is cExpression */
  node_ptr		block;	/* cArcSon other end is cBlock */
} t_dc_for_loop;

typedef struct {
  t_dc_asg_nd	dc_asg_nd;
} t_dc_sql_stmt;

typedef struct {
  t_dc_asg_nd	dc_asg_nd;
} t_dc_comment;

typedef struct {
  t_dc_comment	comment;
} t_dc_java_doc_comment;

typedef struct {
  t_dc_asg_nd	dc_asg_nd;
  node_ptr	except;		/* cArcSon other end is cExpression */
} t_dc_throw_spec;

typedef struct {
  t_dc_asg_nd	dc_asg_nd;
} t_dc_asm_block;

typedef struct {
  gd_graph_common	common;
  int			ordinal;	/* order attribute */
  node_ptr		src;
  node_ptr		dest;
  node_ptr		next;		/* used to link those edges common
					   to a src node for quick traversal */
} t_dc_arc;

typedef struct {
  t_dc_arc		arc;
} t_dc_arc_reference;

typedef struct {
  t_dc_arc_reference	arc_reference;
} t_dc_instance;

typedef struct {
  t_dc_arc_reference	arc_reference;
  t_visb		visb;
  unsigned		virt: 1;
} t_dc_inheritance;

typedef struct {
  t_dc_arc_reference	arc_reference;
} t_dc_implements;

typedef struct {
  t_dc_arc_reference	arc_reference;
} t_dc_arc_friend;

typedef struct {
  t_dc_arc_reference	arc_reference;
} t_dc_arc_def_loc;

typedef struct {
  t_dc_arc_reference	arc_reference;
} t_dc_refers_to;

typedef struct {
  t_dc_arc		arc;
} t_dc_arc_syntax;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_arc_son;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_arc_true;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_arc_false;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_arc_opd;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_import;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_arc_cond;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_arc_for_init;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_arc_for_incr;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_arc_init_val;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_arc_bit_dim;

typedef struct {
  t_dc_arc_syntax	arc_syntax;
} t_dc_arc_array_dim;

/* the union definition of all gd_graph node types */

typedef union {
  gd_graph_common common; /* first field of all node types */

  /*********************
   * gcc types
   *********************/

  /* decl nodes */
  t_gcc_decl		decl;			/* generic version */
  t_namespace_decl	namespace_decl;
  t_function_decl	function_decl;
  t_var_decl		var_decl;
  t_parm_decl		parm_decl;
  t_type_decl		type_decl;
  t_label_decl		label_decl;
  t_const_decl		const_decl;
  t_result_decl		result_decl;
  t_field_decl		field_decl;
  t_template_decl	template_decl;
  t_using_decl 		using_decl;

  /* type nodes */
  t_gcc_type 		type;			/* generic version */
  t_function_type	function_type;
  t_integer_type	integer_type;
  t_array_type		array_type;
  t_void_type		void_type;
  t_real_type		real_type;
  t_complex_type	complex_type;
  t_vector_type		vector_type;
  t_enumeral_type	enumeral_type;
  t_boolean_type	boolean_type;
  t_char_type		char_type;
  t_pointer_type	pointer_type;
  t_offset_type		offset_type;
  t_reference_type	reference_type;
  t_method_type		method_type;
  t_file_type		file_type;
  t_set_type		set_type;
  t_record_type		record_type;
  t_union_type		union_type;
  t_qual_union_type	qual_union_type;
  t_lang_type		lang_type;
  t_template_type_parm	template_type_parm;
  t_template_parm_index	template_parm_index;
  t_template_template_parm	template_template_parm;
  t_bound_template_template_parm	bound_template_template_parm;
  t_typename_type	typename_type;
  t_typeof_type		typeof_type;

  /* generic expression nodes used for statement and expression */
  t_gcc_expr		expr;			/* generic expresion	*/
  t_gcc_expr1		expr1;			/* code_class '1'	*/
  t_gcc_expr2		expr2;			/* code_class '2'	*/
  t_gcc_expr_ineq	expr_ineq;		/* code_class '<'	*/

  /* statment nodes */
  t_block		block;
  t_compound_stmt	compound_stmt;
  t_expr_stmt		expr_stmt;
  t_decl_stmt		decl_stmt;
  t_scope_stmt		scope_stmt;
  t_return_stmt		return_stmt;
  t_stmt_expr		stmt_expr;
  t_break_stmt		break_stmt;
  t_continue_stmt	continue_stmt;
  t_if_stmt		if_stmt;
  t_switch_stmt		switch_stmt;
  t_label_stmt		label_stmt;
  t_try_block		try_block;
  t_handler		handler;
  t_start_catch_stmt	start_catch_stmt;
  t_cleanup_stmt	cleanup_stmt;
  t_case_label		case_label;
  t_goto_stmt		goto_stmt;
  t_for_stmt		for_stmt;
  t_while_stmt		while_stmt;
  t_do_stmt		do_stmt;
  t_asm_stmt		asm_stmt;

  /* reference nodes */
  t_component_ref	component_ref;
  t_bit_field_ref	bit_field_ref;
  t_indirect_ref	indirect_ref;
  t_buffer_ref		buffer_ref;
  t_array_ref		array_ref;
  t_offset_ref		offset_ref;
  t_scope_ref		scope_ref;
  t_member_ref		member_ref;

  /* expression nodes */
  t_constructor		constructor;
  t_compound_expr	compound_expr;
  t_modify_expr		modify_expr;
  t_init_expr		init_expr;
  t_throw_expr		throw_expr;
  t_target_expr		target_expr;
  t_cond_expr		cond_expr;
  t_bind_expr		bind_expr;
  t_call_expr		call_expr;
  t_method_call_expr	method_call_expr;
  t_with_cleanup_expr	with_cleanup_expr;
  t_cleanup_point_expr	cleanup_point_expr;
  t_placeholder_expr	placeholder_expr;
  t_with_record_expr	with_record_expr;
  t_plus_expr		plus_expr;
  t_minus_expr		minus_expr;
  t_mult_expr		mult_expr;
  t_trunc_div_expr	trunc_div_expr;
  t_ceil_div_expr	ceil_div_expr;
  t_floor_div_expr	floor_div_expr;
  t_round_div_expr	round_div_expr;
  t_trunc_mod_expr	trunc_mod_expr;
  t_ceil_mod_expr	ceil_mod_expr;
  t_floor_mod_expr	floor_mod_expr;
  t_round_mod_expr	round_mod_expr;
  t_rdiv_expr		rdiv_expr;
  t_exact_div_expr	exact_div_expr;
  t_fix_trunc_expr	fix_trunc_expr;
  t_fix_ceil_expr	fix_ceil_expr;
  t_fix_floor_expr	fix_floor_expr;
  t_fix_round_expr	fix_round_expr;
  t_float_expr		float_expr;
  t_expon_expr		expon_expr;
  t_negate_expr		negate_expr;
  t_min_expr		min_expr;
  t_max_expr		max_expr;
  t_abs_expr		abs_expr;
  t_ffs_expr		ffs_expr;
  t_lshift_expr		lshift_expr;
  t_rshift_expr		rshift_expr;
  t_lrotate_expr	lrotate_expr;
  t_rrotate_expr	rrotate_expr;
  t_bit_ior_expr	bit_ior_expr;
  t_bit_xor_expr	bit_xor_expr;
  t_bit_and_expr	bit_and_expr;
  t_bit_andtc_expr	bit_andtc_expr;
  t_bit_not_expr	bit_not_expr;
  t_truth_andif_expr	truth_andif_expr;
  t_truth_orif_expr	truth_orif_expr;
  t_truth_and_expr	truth_and_expr;
  t_truth_or_expr	truth_or_expr;
  t_truth_xor_expr	truth_xor_expr;
  t_truth_not_expr	truth_not_expr;
  t_lt_expr		lt_expr;
  t_le_expr		le_expr;
  t_gt_expr		gt_expr;
  t_ge_expr		ge_expr;
  t_eq_expr		eq_expr;
  t_ne_expr		ne_expr;
  t_unordered_expr	unordered_expr;
  t_ordered_expr	ordered_expr;
  t_unlt_expr		unlt_expr;
  t_unle_expr		unle_expr;
  t_ungt_expr		ungt_expr;
  t_unge_expr		unge_expr;
  t_uneq_expr		uneq_expr;
  t_in_expr		in_expr;
  t_set_le_expr		set_le_expr;
  t_card_expr		card_expr;
  t_range_expr		range_expr;
  t_convert_expr	convert_expr;
  t_nop_expr		nop_expr;
  t_non_lvalue_expr	non_lvalue_expr;
  t_save_expr		save_expr;
  t_unsave_expr		unsave_expr;
  t_rtl_expr		rtl_expr;
  t_addr_expr		addr_expr;
  t_reference_expr	reference_expr;
  t_entry_value_expr	entry_value_expr;
  t_complex_expr	complex_expr;
  t_conj_expr		conj_expr;
  t_realpart_expr	realpart_expr;
  t_predecrement_expr	predecrement_expr;
  t_preincrement_expr	preincrement_expr;
  t_postdecrement_expr	postdecrement_expr;
  t_postincrement_expr	postincrement_expr;
  t_va_arg_expr		va_arg_expr;
  t_try_catch_expr	try_catch_expr;
  t_try_finally_expr	try_finally_expr;
  t_goto_subroutine_expr	goto_subroutine_expr;
  t_popdhc_expr		popdhc_expr;
  t_popdcc_expr		popdcc_expr;
  t_label_expr		label_expr;
  t_goto_expr		goto_expr;
  t_return_expr		return_expr;
  t_exit_expr		exit_expr;
  t_loop_expr		loop_expr;
  t_labeled_block_expr	labeled_block_expr;
  t_exit_block_expr	exit_block_expr;
  t_expr_with_file_location	expr_with_file_location;
  t_switch_expr		switch_expr;
  t_sizeof_expr		sizeof_expr;
  t_arrow_expr		arrow_expr;
  t_alignof_expr	alignof_expr;
  t_new_expr		new_expr;
  t_vec_new_expr	vec_new_expr;
  t_delete_expr		delete_expr;
  t_vec_delete_expr	vec_delete_expr;
  t_type_expr		type_expr;
  t_aggr_init_expr	aggr_init_expr;
  t_empty_class_expr	empty_class_expr;
  t_template_id_expr	template_id_expr;
  t_lookup_expr		lookup_expr;
  t_function_name	function_name;
  t_subobject		subobject;
  t_ctor_stmt		ctor_stmt;
  t_ctor_initializer	ctor_initializer;
  t_return_init		return_init;
  t_tag_defn		tag_defn;
  t_identity_conv	identity_conv;
  t_lvalue_conv		lvalue_conv;
  t_qual_conv		qual_conv;
  t_std_conv		std_conv;
  t_ptr_conv		ptr_conv;
  t_pmem_conv		pmem_conv;
  t_base_conv		base_conv;
  t_ref_bind		ref_bind;
  t_user_conv		user_conv;
  t_ambig_conv		ambig_conv;
  t_rvalue_conv		rvalue_conv;
  

  /* constant nodes */
  t_gcc_const		gcc_cst; 	/* generic */
  t_integer_cst		integer_cst;
  t_string_cst		string_cst;
  t_complex_cst		complex_cst;
  t_real_cst		real_cst;
  t_ptrmem_cst		ptrmem_cst;

  /* misc */
  t_op_identifier	op_identifier;
  t_identifier_node	identifier_node;
  t_tree_list		tree_list;
  t_tree_vec		tree_vec;
  t_error_mark		error_mark;
  t_default_arg		default_arg;
  t_cplus_binding	cplus_binding;
  t_overload		overload;
  t_wrapper		wrapper;

  /***********************
   * Datrix Nodes
   ***********************/

  t_dc_asg_nd			dc_asg_nd;

  t_dc_identifier		dc_identifier;

  t_dc_type			dc_type;
  t_dc_built_in_type		dc_built_in_type;
  t_dc_aggr_type		dc_aggr_type;
  t_dc_enum_type		dc_enum_type;
  t_dc_templ_parm_type		dc_templ_parm_type;
  t_dc_generic_type		dc_generic_type;
  t_dc_generated_type		dc_generated_type;
  t_dc_alias_type		dc_alias_type;
  t_dc_forward_type		dc_forward_type;
  t_dc_array_type		dc_array_type;
  t_dc_ptr_type			dc_ptr_type;
  t_dc_ref_type			dc_ref_type;
  t_dc_fct_type			dc_fct_type;

  t_dc_scope			dc_scope;
  t_dc_system			dc_system;
  t_dc_scope_glb		dc_scope_glb;
  t_dc_scope_compile		dc_scope_compile;
  t_dc_scope_file		dc_scope_file;
  t_dc_package			dc_package;
  t_dc_name_space		dc_name_space;
  t_dc_block			dc_block;
  t_dc_static_block		dc_static_block;
  t_dc_try_block		dc_try_block;
  t_dc_catch_block		dc_catch_block;
  t_dc_finally_block		dc_finally_block;
  t_dc_synchronized_block	dc_synchronized_block;


  t_dc_object			dc_object;
  t_dc_formal_fct_param		dc_formal_fct_param;
  t_dc_formal_catch_param	dc_formal_catch_param;

  t_dc_function			dc_function;

  t_dc_label			dc_label;
  t_dc_case_label		dc_case_label;
  t_dc_default_label		dc_default_label;

  t_dc_enumerator		dc_enumerator;

  t_dc_expression		dc_expression;
  t_dc_empty_expr		dc_empty_expr;
  t_dc_literal			dc_literal;
  t_dc_syntax_error		dc_syntax_error;
  t_dc_name_ref			dc_name_ref;
  t_dc_operator			dc_operator;
  t_dc_unary_operator		dc_unary_operator;
  t_dc_binary_operator		dc_binary_operator;
  t_dc_question_operator	dc_question_operator;
  t_dc_init_list		dc_init_list;
  t_dc_actual_fct_param_list	dc_actual_fct_param_list;
  t_dc_actual_templ_param_list	dc_actual_templ_param_list;
  t_dc_creation_expr		dc_creation_expr;
  t_dc_destroy_expr		dc_destroy_expr;

  t_dc_ctrl_stmt		dc_ctrl_stmt;
  t_dc_selection		dc_selection;
  t_dc_if			dc_if;
  t_dc_switch			dc_switch;
  t_dc_jump			dc_jump;
  t_dc_goto			dc_goto;
  t_dc_return			dc_return;
  t_dc_break			dc_break;
  t_dc_continue			dc_continue;

  t_dc_loop			dc_loop;
  t_dc_do_while_loop		dc_do_while_loop;
  t_dc_while_do_loop		dc_while_do_loop;
  t_dc_for_loop			dc_for_loop;

  t_dc_sql_stmt			dc_sql_stmt;
  t_dc_comment			dc_comment;
  t_dc_java_doc_comment		dc_java_doc_comment;

  t_dc_throw_spec		dc_throw_spec;
  t_dc_asm_block		dc_asm_block;
 
  t_dc_arc			dc_arc;
  t_dc_arc_reference		dc_arc_reference;
  t_dc_instance			dc_instance;
  t_dc_inheritance		dc_inheritance;
  t_dc_implements		dc_implements;
  t_dc_arc_friend		dc_arc_friend;
  t_dc_arc_def_loc		dc_arc_def_loc;
  t_dc_refers_to		dc_refers_to;

  t_dc_arc_syntax		dc_arc_syntax;
  t_dc_arc_son			dc_arc_son;
  t_dc_arc_true			dc_arc_true;
  t_dc_arc_false		dc_arc_false;
  t_dc_arc_opd			dc_arc_opd;
  t_dc_import			dc_import;
  t_dc_arc_cond			dc_arc_cond;
  t_dc_arc_for_init		dc_arc_for_init;
  t_dc_arc_for_incr		dc_arc_for_incr;
  t_dc_arc_init_val		dc_arc_init_val;
  t_dc_arc_bit_dim		dc_arc_bit_dim;
  t_dc_arc_array_dim		dc_arc_array_dim;

} gd_graph_node;

/* struct gd_graph, the header node used to refer to the gd_graph.
 * It combines the string table, hash table and gd_graph table
 * into single structure.
 * 
 * Notes:
 *   - all pointers within structures are integers, this lets
 *     us read and write it to files trivially
 *   - character string references are indexes directly into
 *     the string buffer. The hash table is only used for
 *     insertion of strings, so that a string value is only
 *     stored once.
 */

typedef struct {
  /* string table stuff */
  unsigned long stringBuffLen;
  unsigned long curStringPos;
  char * stringBuffer;

  /* hash table */
  /* hash table only stores location in string
   * table. Therefore, no need for a complex structure */
  unsigned long hashTableLen;
  long *hashTable;

  /* node table */
  unsigned long nodeBuffLen;
  unsigned long curNodePos;
  node_ptr	firstNode;
  gd_graph_node * nodeBuffer;
} gd_graph;

/******************
 * construct/deconstruct
 ******************/

/* allocate_gd_graph - build an empty gd_graph with the given sizes */

gd_graph allocate_gd_graph(unsigned long StringSize, unsigned long HashSize,
    			unsigned long GraphSize);

void deallocate_gd_graph(gd_graph g);

/******************
 * file ops
 ******************/

/* gd_graph_write_binary_file -- write the gd_graph to a binary file */
/* 1 means success, 0 means failure */
int gd_graph_write_binary_file(const char * filename, gd_graph g);

/* gd_graph_read_binary - read the gd_graph from a binary file. Since only
 * the used gd_graph is stored as a binary file, there is no extra space
 * the three extra parameters are multipliers to expand each of the
 * tables.  For example, if no new string tokens will be generated,
 * (only new nodes), then set the string_mult and has_mult to 1.0 */

/* TODO - hash table adjustment (hash_mult > 1.0) not done yet */

gd_graph gd_graph_read_binary_file(const char * filename, float string_mult, 
				  float hash_mult, float node_mult);


/******************
 * string ops
 ******************/

/* lookup_string - returns -1 if the string is not in the table,
 *     or the index of the string in the string table if there */

long gd_lookup_string(gd_graph *g, const char * string);

/* add_string - adds the string to the table and returns the
 *    index. If the string is already there, then the existing
 *    index is returned, -1 means error inserting string */

long gd_add_string(gd_graph *g, const char * newstring);

/***********************
 * Storage Management
 ***********************/

#define DEFAULT_GROW_NODES 2048
#define DEFAULT_GROW_CHARS 2048
#define DEFAULT_GROW_HASH  1023

int gd_grow_nodes(gd_graph * g, unsigned long grow_amt);

#endif /* _GD_GRAPH_H_ */
